from .rest_client import RestClient


class PublicClient(RestClient):
  """zcoinbase public client API.

  Public API Documentation:
    https://docs.pro.coinbase.com/#market-data

  Attributes:
    rest_url: REST API URL. Defaults to https://api.pro.coinbase.com
  """

  SANDBOX_URL = 'https://api-public.sandbox.pro.coinbase.com'
  PROD_URL = 'https://api.pro.coinbase.com'

  def __init__(self, rest_url=PROD_URL):
    super().__init__(rest_url)

  @classmethod
  def make_prod_client(cls):
    return cls(rest_url=cls.PROD_URL)

  @classmethod
  def make_sandbox_client(cls):
    return cls(rest_url=cls.SANDBOX_URL)

  def get_products(self):
    return self._send_get('products')

  def get_product(self, product_id):
    return self._send_get('products/{}'.format(product_id))

  def get_order_book(self, product_id, level=1):
    if 1 <= level >= 3:
      raise ValueError('Level must be between 1 and three.')
    params = {'level': level}
    return self._send_get('products/{}/book'.format(product_id), params=params)

  def get_ticker(self, product_id):
    return self._send_get('products/{}/ticker'.format(product_id))

  def get_trades(self, product_id):
    return self._send_paginated_get('products/{}/trades')

  def get_historic_rates(self, product_id, start=None, end=None, granularity=None):
    params = dict()
    accepted_granularities = {60, 300, 900, 3600, 21600, 86400}
    if granularity is not None:
      if granularity not in accepted_granularities:
        raise ValueError(
          'Specified Granularity is {}, must be in approved values: {}'.format(granularity, accepted_granularities))
      params['granularity'] = granularity
    if start is not None:
      params['start'] = start
    if end is not None:
      params['end'] = end
    return self._send_get('products/{}/candles'.format(product_id), params=params)

  def get_24hr_stats(self, product_id):
    return self._send_get('products/{}/stats'.format(product_id))

  def get_currencies(self):
    return self._send_get('currencies')

  def get_currency(self, currency_id):
    return self._send_get('currencies/{}'.format(currency_id))

  def get_time(self):
    return self._send_get('time')



import hmac
import hashlib
import time
import base64

from requests.auth import AuthBase


class CoinbaseAuth(AuthBase):
  """Coinbase Auth.

  API Information: https://docs.pro.coinbase.com/#signing-a-message
  """

  def __init__(self, api_key, api_secret, passphrase,
               time_provider=lambda: time.time()):
    self.api_key = api_key
    self.api_secret = api_secret
    self.passphrase = passphrase
    self.time_provider = time_provider

  def __call__(self, request):
    timestamp = str(self.time_provider())
    message = ''.join([timestamp, request.method, request.path_url, (request.body or '')])
    request.headers.update(
      CoinbaseAuth.get_auth_headers(timestamp, message, self.api_key, self.api_secret, self.passphrase))
    return request

  @staticmethod
  def get_websocket_verification(api_key, api_secret, passphrase):
    timestamp = str(time.time())
    message = ''.join([timestamp, 'GET', '/users/self/verify'])
    auth = CoinbaseAuth.make_auth(timestamp, message, api_key, api_secret, passphrase)
    return {
      'signature': auth[0],
      'timestamp': auth[1],
      'key': auth[2],
      'passphrase': auth[3]
    }

  @staticmethod
  def get_auth_headers(timestamp, message, api_key, api_secret, passphrase):
    auth = CoinbaseAuth.make_auth(timestamp, message, api_key, api_secret, passphrase)
    return {
      'Content-Type': 'Application/JSON',
      'CB-ACCESS-SIGN': auth[0],
      'CB-ACCESS-TIMESTAMP': auth[1],
      'CB-ACCESS-KEY': auth[2],
      'CB-ACCESS-PASSPHRASE': auth[3]
    }

  @staticmethod
  def make_auth(timestamp, message, api_key, api_secret, passphrase):
    message = message.encode('ascii')
    hmac_key = base64.b64decode(api_secret)
    signature = hmac.new(hmac_key, message, hashlib.sha256)
    signature_b64 = base64.b64encode(signature.digest()).decode('utf-8')
    return signature_b64, timestamp, api_key, passphrase




from enum import Enum, auto


class LogLevel(Enum):
  """Enum for LogLevels in zcoinbase, pick anything from no logging to VERBOSE logging."""
  NO_LOG = 0  # Don't log anything.
  ERROR_LOG = 1  # Only errors logged.
  BASIC_MESSAGES = 2  # Errors logged as well as connection and disconnection.
  VERBOSE_LOG = 3  # Log Everything that might be interesting.

  def __eq__(self, other):
    if self.__class__ is other.__class__:
      return self.value == other.value
    return NotImplemented

  def __gt__(self, other):
    return self.value > other.value

  def __ge__(self, other):
    return self.value >= other.value

  def __lt__(self, other):
    return self.value < other.value

  def __le__(self, other):
    return self.value <= other.value


class OrderSide(Enum):
  BUY = auto()
  SELL = auto()


class TimeInForce(Enum):
  GOOD_TILL_CANCEL = 'GTC'
  GOOD_UNTIL_TIME = 'GTT'
  IMMEDIATE_OR_CANCEL = 'IOC'
  FILL_OR_KILL = 'FOK'


class Stop(Enum):
  NONE = 'none'
  LOSS = 'loss'
  ENTRY = 'entry'


class SelfTradePrevention(Enum):
  DECREASE_AND_CANCEL = 'dc'
  CANCEL_OLDEST = 'co'
  CANCEL_NEWEST = 'cn'
  CANCEL_BOTH = 'cb'


class OrderStatus(Enum):
  OPEN = 'open'
  PENDING = 'pending'
  ACTIVE = 'active'


class TransferType(Enum):
  DEPOSIT = 'deposit'
  WITHDRAW = 'withdraw'


class ReportType(Enum):
  FILLS = 'fills'
  ACCOUNT = 'account'


class ReportFormat(Enum):
  PDF = 'pdf'
  CSV = 'csv'



import requests
import json


class RestClient:
  def __init__(self, rest_url='https://api.pro.coinbase.com'):
    self.rest_url = rest_url.rstrip('/')
    self.session = requests.sessions.Session()
    self.auth = None

  def _send_get(self, endpoint, params=None):
    url = '{}/{}'.format(self.rest_url, endpoint)
    result = self.session.get(url, params=params, auth=self.auth)
    if result.status_code != 200:
      raise RuntimeError('ErrorCode: {} Message: {}\nGET Request to {} w/ params {} FAILED'.format(result.status_code,
                                                                                                   result.json()[
                                                                                                     'message'], url,
                                                                                                   params))
    return result.json()

  def _send_paginated_get(self, endpoint, params=None):
    if params is None:
      params = dict()
    url = '{}/{}'.format(self.rest_url, endpoint)
    while True:
      r = self.session.get(url, params=params, auth=self.auth, timeout=30)
      results = r.json()
      for result in results:
        yield result
      if not r.headers.get('cb-after') or params.get('before') is not None:
        break
      else:
        params['after'] = r.headers['cb-after']

  def _send_post(self, endpoint, params=None, data=None):
    return RestClient._append_status_code(
      self.session.post('{}/{}'.format(self.rest_url, endpoint),
                        data=json.dumps(params) if params is not None else None, auth=self.auth))

  def _send_delete(self, endpoint, params=None):
    return RestClient._append_status_code(
      self.session.delete('{}/{}'.format(self.rest_url, endpoint), params=params, auth=self.auth))

  @staticmethod
  def _append_status_code(response):
    response_json = response.json()
    response_json['http_code'] = response.status_code
    return response_json



from zcoinbase.websocket_client import CoinbaseWebsocket
from zcoinbase.util import OrderSide, TimeInForce, SelfTradePrevention, Stop, OrderStatus, TransferType, ReportType, \
  ReportFormat, LogLevel
from zcoinbase.public_client import PublicClient
from zcoinbase.authenticated_client import AuthenticatedClient
from zcoinbase.coinbase_order_book import CoinbaseOrderBook, ProductOrderBook
from zcoinbase.historical_data_downloader import HistoricalDownloader



import json
import uuid

from .coinbase_auth import CoinbaseAuth
from .public_client import PublicClient
from .util import OrderSide, TimeInForce, SelfTradePrevention, Stop, OrderStatus, TransferType, ReportType, ReportFormat


class AuthenticatedClient(PublicClient):
  def __init__(self, api_key, api_secret, passphrase,
               rest_url='https://api.pro.coinbase.com'):
    super().__init__(rest_url)
    self.auth = CoinbaseAuth(api_key, api_secret, passphrase)

  def get_all_accounts(self):
    """List all accounts.

    https://docs.pro.coinbase.com/#list-accounts
    """
    return self._send_get('accounts')

  def get_account(self, account_id):
    """Information for a single account.

    https://docs.pro.coinbase.com/#get-an-account
    """
    return self._send_get('accounts/{}'.format(account_id))

  def get_account_history(self, account_id):
    """Gets your account history.

    https://docs.pro.coinbase.com/#get-account-history
    """
    return self._send_paginated_get(self, 'accounts/{}/ledger'.format(account_id))

  def get_account_holds(self, account_id):
    """Gets holds on your account.

    https://docs.pro.coinbase.com/#get-holds
    """
    return self._send_paginated_get(self, 'accounts/{}/holds'.format(account_id))

  @staticmethod
  def make_order_uuid():
    return str(uuid.uuid4())

  def limit_order(self, side: OrderSide, product_id, price, size,
                  time_in_force=TimeInForce.GOOD_TILL_CANCEL,
                  cancel_after=None, post_only=None,
                  self_trade_prevention=SelfTradePrevention.DECREASE_AND_CANCEL,
                  stop=Stop.NONE, stop_price=None):
    """Places a limit order.

    https://docs.pro.coinbase.com/#place-a-new-order
    """
    if time_in_force is TimeInForce.GOOD_UNTIL_TIME and cancel_after is None:
      raise ValueError('cancel_after is required with GOOD_UNTIL_TIME')
    if cancel_after is not None and time_in_force is not TimeInForce.GOOD_UNTIL_TIME:
      raise ValueError('cannot set cancel after for orders other than GOOD_UNTIL_TIME')
    if post_only is not None and \
        (time_in_force is TimeInForce.IMMEDIATE_OR_CANCEL or time_in_force is TimeInForce.FILL_OR_KILL):
      raise ValueError('cannot set post_only with IMMEDIATE_OR_CANCEL or FILL_OR_KILL')
    if stop is not Stop.NONE and stop_price is None:
      raise ValueError('stop_price is required with LOSS or ENTRY')

    params = {
      'type': 'limit',
      'client_oid': AuthenticatedClient.make_order_uuid(),
      'side': 'buy' if side == OrderSide.BUY else 'sell',
      'price': price,
      'size': size,
      'product_id': product_id,
      'time_in_force': time_in_force.value,
      'stp': self_trade_prevention.value
    }
    if cancel_after is not None:
      params['cancel_after'] = cancel_after
    if post_only is not None:
      params['post_only'] = post_only
    if stop is not Stop.NONE:
      params['stop'] = stop.value
      params['stop_price'] = stop_price
    response = self._send_post('orders', params=params)
    response['client_oid'] = params['client_oid']
    return response

  def market_order(self, side: OrderSide, product_id, size=None, funds=None,
                   self_trade_prevention=SelfTradePrevention.DECREASE_AND_CANCEL,
                   stop=Stop.NONE, stop_price=None):
    """Places a market Order.

    https://docs.pro.coinbase.com/#place-a-new-order
    """
    if size is None and funds is None:
      raise ValueError('must specify size or funds for market order.')
    if stop is not Stop.NONE and stop_price is None:
      raise ValueError('stop_price is required with LOSS or ENTRY')
    params = {
      'type': 'market',
      'client_oid': AuthenticatedClient.make_order_uuid(),
      'side': 'buy' if side == OrderSide.BUY else 'sell',
      'product_id': product_id,
      'stp': self_trade_prevention.value,
    }
    if size is not None:
      params['size'] = size
    if funds is not None:
      params['funds'] = funds
    if stop is not Stop.NONE:
      params['stop'] = stop.value
      params['stop_price'] = stop_price
    response = self._send_post('orders', params=params)
    response['client_oid'] = params['client_oid']
    return response

  def cancel_order(self, order_id, is_client_oid=False, product_id=None):
    """Cancels a single order.

    https://docs.pro.coinbase.com/#cancel-an-order
    """
    params = {}
    if product_id is not None:
      params['product_id'] = product_id
    request_string = 'orders/client:{}' if is_client_oid else 'orders/{}'
    return self._send_delete(request_string.format(order_id), params=params)

  def cancel_all_orders(self, product_id=None):
    """Cancels all orders.

    https://docs.pro.coinbase.com/#cancel-all
    """
    params = {}
    if product_id is not None:
      params['product_id'] = product_id
    return self._send_delete('orders', params=params)

  def list_orders(self, status: list[OrderStatus], product_id=None):
    """Lists Current Open Orders.

    https://docs.pro.coinbase.com/#list-orders
    """
    params = {}
    if product_id is not None:
      params['product_id'] = product_id
    for s in status:
      try:
        params['status'].append(s.value)
      except KeyError:
        params['status'] = [s.value]
    return self._send_paginated_get('orders', params=params)

  def get_order(self, order_id, is_client_oid=False):
    """Gets a single order by ID.

    https://docs.pro.coinbase.com/#get-an-order
    """
    request_string = 'orders/client:{}' if is_client_oid else 'orders/{}'
    return self._send_get(request_string.format(order_id))

  def list_fills(self, order_id=None, product_id=None):
    """List Fills.

    https://docs.pro.coinbase.com/#list-fills
    """
    if order_id is None and product_id is None:
      raise ValueError('order_id or product_id is required')
    params = {}
    if order_id is not None:
      params['order_id'] = order_id
    if product_id is not None:
      params['product_id'] = product_id
    return self._send_paginated_get('fills', params=params)

  def get_current_exchange_limits(self):
    """Gets current exchange limits.

    https://docs.pro.coinbase.com/#get-current-exchange-limits
    """
    return self._send_get('users/self/exchange-limits')

  def list_transfers(self, transfer_type: TransferType, profile_id=None, before=None, after=None, limit=None):
    """List Transfers.

    Deposits: https://docs.pro.coinbase.com/#list-deposits
    Withdrawals: https://docs.pro.coinbase.com/#list-withdrawals
    """
    if limit is not None and limit > 100:
      raise ValueError('cannot request more than 100 transfers')
    params = {
      'type': transfer_type.value
    }
    if limit is not None:
      params['limit'] = limit
    if profile_id is not None:
      params['profile_id'] = profile_id
    if before is not None:
      params['before'] = before
    if after is not None:
      params['after'] = after
    return self._send_get('transfers', params=params)

  def list_deposits(self, **kwargs):
    """List Deposits.

    https://docs.pro.coinbase.com/#list-deposits
    """
    return self.list_transfers(transfer_type=TransferType.DEPOSIT, **kwargs)

  def list_withdrawals(self, **kwargs):
    """List Withdrawals.

    https://docs.pro.coinbase.com/#list-withdrawals
    """
    return self.list_transfers(transfer_type=TransferType.WITHDRAW, **kwargs)

  # There is a Bug in the Coinbase API, so this doesn't work.
  # def get_transfer(self, transfer_id):
  #   return self._send_get('transfers/:{}'.format(transfer_id))

  def get_payment_methods(self):
    """List your Payment Methods.

    https://docs.pro.coinbase.com/#list-payment-methods
    """
    return self._send_get('payment-methods')

  def deposit_funds_payment(self, amount, currency, payment_method_id):
    """Deposit funds from a payment method.

    https://docs.pro.coinbase.com/#payment-method
    """
    params = {
      'amount': amount,
      'currency': currency,
      'payment_method_id': payment_method_id
    }
    return self._send_post('deposits/payment-method', params=params)

  def deposit_funds_coinbase(self, amount, currency, coinbase_account_id):
    """Deposit funds from a Coinbase Account.

    https://docs.pro.coinbase.com/#coinbase
    """
    params = {
      'amount': amount,
      'currency': currency,
      'coinbase_account_id': coinbase_account_id
    }
    return self._send_post('deposits/coinbase-account', params=params)

  def generate_crypto_deposit_address(self, coinbase_account_id):
    """Generate an address for crypto deposits.

    https://docs.pro.coinbase.com/#generate-a-crypto-deposit-address
    """
    return self._send_post('coinbase-accounts/{}/addresses'.format(coinbase_account_id))

  def list_coinbase_accounts(self):
    """Get a list of coinbase accounts.

    https://docs.pro.coinbase.com/#list-payment-methods
    """
    return self._send_get('coinbase-accounts')

  def get_fees(self):
    """Get active fees for your account.

    https://docs.pro.coinbase.com/#fees
    """
    return self._send_get('fees')

  def generate_report(self, report_type: ReportType, start_date, end_date,
                      product_id=None, account_id=None, report_format: ReportFormat = ReportFormat.PDF,
                      email=None):
    """Generates a report.

    https://docs.pro.coinbase.com/#create-a-new-report
    """
    if report_type is ReportType.FILLS and product_id is None:
      raise ValueError('product id is required for ReportType.FILLS')
    if report_type is ReportType.ACCOUNT and account_id is None:
      raise ValueError('account id is required for ReportType.ACCOUNT')
    params = {
      'type': report_type.value,
      'start_date': start_date,
      'end_date': end_date,
      'report_format': report_format.value
    }
    if product_id is not None:
      params['product_id'] = product_id
    if account_id is not None:
      params['account_id'] = account_id
    if email is not None:
      params['email'] = email
    return self._send_post('reports', params=params)

  def get_report_status(self, report_id):
    """Get the status of a report given the id.

    https://docs.pro.coinbase.com/#get-report-status
    """
    return self._send_get('reports/{}'.format(report_id))

  def get_profiles(self, active=False):
    """Get Profiles.

    https://docs.pro.coinbase.com/#list-profiles
    """
    params = {'active': active}
    return self._send_get('profiles', params=params)

  def get_single_profile(self, profile_id):
    """Get a Profile.

    https://docs.pro.coinbase.com/#get-a-profile
    """
    return self._send_get('profiles/{}'.format(profile_id))

  def create_profile_transfer(self, from_profile, to_profile, currency, amount):
    """Creates a transfer from one profile to another.

    https://docs.pro.coinbase.com/#create-profile-transfer
    """
    params = {
      'from': from_profile,
      'to': to_profile,
      'currency': currency,
      'amount': amount
    }
    return self._send_post('profiles/transfer', params=params)

  def get_trailing_volume(self):
    """Gets 30-day trailing volume for all products.

    https://docs.pro.coinbase.com/#trailing-volume
    """
    return self._send_get('users/self/trailing-volume')

  def get_margin_profile_information(self, product_id):
    """Get margin profile information.

    https://docs.pro.coinbase.com/#get-margin-profile-information
    """
    params = {'product_id': product_id}
    return self._send_get('margin/profile_information', params=params)

  def get_buying_power(self, product_id):
    """Get's buying/selling power of particular product.

    https://docs.pro.coinbase.com/#get-buying-power
    """
    params = {'product_id': product_id}
    return self._send_get('margin/buying_power', params=params)

  def get_withdrawl_power(self, currency):
    """Returns the max amount of given currency you can withdraw from your margin profile.

    https://docs.pro.coinbase.com/#get-withdrawal-power
    """
    params = {'currency': currency}
    return self._send_get('margin/withdrawal_power', params=params)

  def get_all_withdrawal_powers(self, currency):
    """Returns max amount of each currency you can withdraw from your margin profile.

    https://docs.pro.coinbase.com/#get-all-withdrawal-powers
    """
    return self._send_get('margin/withdrawal_power_all')

  def get_exit_plan(self):
    """Returns a liquidation strategy to get your equity percentage back to acceptable level.

    https://docs.pro.coinbase.com/#get-exit-plan
    """
    return self._send_get('margin/exit_plan')

  def get_liquidation_history(self):
    """Returns a list of liquidations performed to get your equity percentage about to acceptable level.

    https://docs.pro.coinbase.com/#list-liquidation-history
    """
    return self._send_get('margin/liquidation_history')

  def get_position_refresh_amounts(self):
    """Returns amount in USD of loads that will be renewed in the next day and the day after.

    https://docs.pro.coinbase.com/#get-position-refresh-amounts
    """
    return self._send_get('margin/position_refresh_amounts')

  def get_margin_status(self):
    """Returns margin status.

    https://docs.pro.coinbase.com/#get-margin-status
    """
    return self._send_get('margin/status')

  def get_oracle(self):
    """Get cryptographically signed prices read fo be posted on-chain using Open-Oracle smart contracts.

    https://docs.pro.coinbase.com/#oracle
    """
    return self._send_get('oracle')



# Maintains a level2 order book of Coinbase

from operator import neg
from sortedcontainers import SortedDict
from threading import Lock
from typing import Text

from zcoinbase import CoinbaseWebsocket


class ProductOrderBook:
  def __init__(self, product_id):
    self.product_id = product_id
    self.asks = SortedDict(lambda key: float(key))
    self.asks_lock = Lock()
    self.bids = SortedDict(lambda key: neg(float(key)))
    self.bids_lock = Lock()
    self.first_bids_lock = Lock()
    self.first_bids_lock.acquire()
    self.first_asks_lock = Lock()
    self.first_asks_lock.acquire()

  def top_n_string(self, n=None):
    """Returns the "Top-N" asks/bids in the order-book in string form.

    Params:
      n: How many of the top
    """
    with self.bids_lock and self.asks_lock:
      return ProductOrderBook._make_formatted_string(
        bids=ProductOrderBook._make_sorted_dict_slice(self.bids, stop=n),
        asks=ProductOrderBook._make_sorted_dict_slice(self.asks, stop=n)
      )

  def get_book(self, top_n=None):
    """Returns the order book as a dict with keys 'asks' and 'bids' and tuples of [price, size].

    Params:
      top_n: The depth of the order book to return.
    """
    return {
      'asks': self.get_asks(top_n=top_n),
      'bids': self.get_bids(top_n=top_n)
    }

  def get_asks(self, top_n=None):
    """Get the 'asks' part of the order book.

    Params:
      top_n: The depth of the order book to return.
    """
    with self.asks_lock:
      return ProductOrderBook._make_slice(self.asks, stop=top_n)

  def get_bids(self, top_n=None):
    """Get the 'bids' part of the order book.

        Params:
          top_n: The depth of the order book to return.
        """
    with self.bids_lock:
      bids_slice = ProductOrderBook._make_slice(self.bids, stop=top_n)

  # Private API Below this Line.
  def _init_bids(self, bids):
    with self.bids_lock:
      for price, size in bids:
        self.bids[price] = float(size)
      self.first_bids_lock.release()

  def _init_asks(self, asks):
    with self.asks_lock:
      for price, size in asks:
        self.asks[price] = float(size)
      self.first_asks_lock.release()

  def _consume_changes(self, changes):
    for side, price, size in changes:
      if side == 'buy':
        self._consume_buy(price, size)
      elif side == 'sell':
        self._consume_sell(price, size)

  def _consume_buy(self, price, size):
    fsize = float(size)
    # Wait for _init_bids to run.
    if self.first_bids_lock.locked():
      self.first_bids_lock.acquire()
      self.first_bids_lock.release()
    with self.bids_lock:
      if str(fsize) == '0.0':
        del self.bids[price]
      else:
        self.bids[price] = fsize

  def _consume_sell(self, price, size):
    fsize = float(size)
    # Wait for _init_asks to run.
    if self.first_asks_lock.locked():
      self.first_asks_lock.acquire()
      self.first_asks_lock.release()
    with self.asks_lock:
      if str(fsize) == '0.0':
        del self.asks[price]
      else:
        self.asks[price] = fsize

  @staticmethod
  def _make_formatted_string(bids, asks):
    overall_format = "BIDS:\n{}\n\nASKS:\n{}\n\n"
    format_str = 'PRICE: {}, SIZE: {}'
    return overall_format.format(
      '\n'.join(format_str.format(str(price), str(bids[price])) for price in bids.keys()),
      '\n'.join(format_str.format(str(price), str(asks[price])) for price in asks.keys()))

  def __repr__(self):
    """Print the entire order book."""
    with self.asks_lock and self.bids_lock:
      return ProductOrderBook._make_formatted_string(self.bids, self.asks)

  @staticmethod
  def _make_sorted_dict_slice(orders: SortedDict, start=None, stop=None):
    return SortedDict(orders.key, [(key, orders[key]) for key in orders.islice(start=start, stop=stop)])

  @staticmethod
  def _make_slice(orders: SortedDict, start=None, stop=None):
    return [(key, orders[key]) for key in orders.islice(start=start, stop=stop)]


class CoinbaseOrderBook:
  def __init__(self, cb_ws: CoinbaseWebsocket):
    self.coinbase_websocket = cb_ws
    self.coinbase_websocket.add_channel('level2')
    self.order_books = {}
    for product in self.coinbase_websocket.products_to_listen:
      self.order_books[product] = ProductOrderBook(product)
    self.coinbase_websocket.add_channel_function('l2update',
                                                 lambda message: self.update_order_book(message['product_id'],
                                                                                        message['changes']),
                                                 refresh_subscriptions=False)
    self.coinbase_websocket.add_channel_function('snapshot',
                                                 lambda message: self.initial_snapshot(message['product_id'],
                                                                                       bids=message['bids'],
                                                                                       asks=message['asks']),
                                                 refresh_subscriptions=False)

  def get_order_book(self, product_id) -> ProductOrderBook:
    if product_id in self.order_books:
      return self.order_books[product_id]
    else:
      raise ValueError('Don\'t have order book for {}'.format(product_id))

  def get_tracked_products(self):
    return self.order_books.keys()

  def add_order_books(self, product_ids: list[Text], refresh_subscriptions=True):
    for product_id in product_ids:
      if product_id not in self.order_books:
        self.order_books[product_id] = ProductOrderBook(product_id)
        self.coinbase_websocket.add_product(product_id, refresh_subscriptions=False)
    if refresh_subscriptions:
      self.coinbase_websocket.subscribe()

  def initial_snapshot(self, product_id, bids, asks):
    if product_id in self.order_books:
      self.order_books[product_id]._init_bids(bids)
      self.order_books[product_id]._init_asks(asks)

  def update_order_book(self, product_id, changes):
    if product_id in self.order_books:
      self.order_books[product_id]._consume_changes(changes)



import csv
import datetime
import math
import sys
import pandas as pd

from dateutil import parser
from functools import partial
from typing import Text, Callable
from zcoinbase import PublicClient
from zcoinbase.internal import RateLimitedExecutionQueue


# Magically use absl logging if available.
import importlib

absl_spec = importlib.util.find_spec('absl')
if absl_spec is not None:
  # Use absl logging if it's available on the environment.
  from absl import logging

  logging.set_verbosity(logging.INFO)
else:
  # Use python default logging if it is not available.
  import logging

# Import progressbar if available.
progressbar_spec = importlib.util.find_spec('progressbar')
if progressbar_spec is not None:
  import progressbar


class HistoricalDownloader:
  TIMESLICE_MAPPINGS = {
    '1m': '60',
    '5m': '300',
    '15m': '900',
    '1h': '3600',
    '6h': '21600',
    '1d': '86400'
  }

  TIMESLICE_MUST_BE_ONE_OF_STRING = ", ".join(
    list(TIMESLICE_MAPPINGS.keys()) + list(TIMESLICE_MAPPINGS.values()))

  _MAX_CANDLES = 300

  _COLUMN_HEADERS = ['time', 'low', 'high', 'open', 'close', 'volume']

  def __init__(self, product_id: Text, start_time, end_time, granularity: Text,
               rest_url=PublicClient.PROD_URL, enable_progressbar=True):
    self.public_client = PublicClient(rest_url=rest_url)
    self.product_id = product_id
    self.enable_progressbar = enable_progressbar
    if isinstance(start_time, str):
      self.start_time = parser.parse(start_time)
    elif isinstance(start_time, datetime.datetime):
      self.start_time = start_time
    else:
      raise ValueError('start_time must be either string or datetime.datetime.')
    if isinstance(end_time, str):
      self.end_time = parser.parse(end_time)
    elif isinstance(end_time, datetime.datetime):
      self.end_time = end_time
    else:
      raise ValueError('end_time must be either string or datetime.datetime')
    if start_time > end_time:
      raise ValueError('start_time must be before end_time')
    if not HistoricalDownloader.validate_granularity(granularity):
      raise ValueError('Granularity must be one of [{}]'.format(HistoricalDownloader.TIMESLICE_MUST_BE_ONE_OF_STRING))
    if granularity in HistoricalDownloader.TIMESLICE_MAPPINGS:
      self.granularity = int(HistoricalDownloader.TIMESLICE_MAPPINGS[granularity])
    else:
      self.granularity = int(granularity)

  def _make_interval_call(self, product_id: Text, start_time: datetime.datetime, end_time: datetime.datetime,
                          granularity):
    return self.public_client.get_historic_rates(product_id, start=start_time.isoformat(), end=end_time.isoformat(),
                                                 granularity=granularity)

  def _download_with_queue(self, row_function: Callable[[list[Text]], None], queue: RateLimitedExecutionQueue):
    """A _download method that allows you to set your own RateLimitedExecutionQueue, might be useful in larger API."""
    # First find the required number of calls at the given granularity, and the start/end times that should be used.
    required_calls = HistoricalDownloader._solve_required_calls(start_time=self.start_time, end_time=self.end_time,
                                                                granularity=self.granularity)
    # Create a time-estimate.
    estimated_time = datetime.timedelta(seconds=(len(required_calls) / queue.max_calls_per_interval) * queue.interval)
    logging.info(
      'Making {} calls to Coinbase API. This will take approximately: {}'.format(len(required_calls), estimated_time))
    results = []
    for start_time, end_time in required_calls:
      logging.debug('Call with Start Time: {} and End Time: {}'.format(start_time, end_time))
      results.append(queue.add_function_to_pool(
        partial(self._make_interval_call, self.product_id, start_time, end_time, self.granularity)))
    bar = None
    bar_progress = 0
    if self.enable_progressbar and 'progressbar' in sys.modules:
      bar = progressbar.ProgressBar(maxval=len(results),
                                    widgets=[progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage(), ' [',
                                             progressbar.ETA(), '] '])
      bar.start()
    for result in results:
      actual_result = sorted(result.get(), key=lambda x: x[0])
      logging.debug('Actual Returned Result: {}'.format(actual_result))
      if bar:
        bar_progress += 1
        bar.update(bar_progress)
      for row in actual_result:
        row_function(row)
    if bar:
      bar.finish()

  def _download(self, row_function: Callable[[list], None]):
    """Downloads historical data and calls the row_function with the text of each row."""
    # Set up RateLimitedExecutionQueue with max_calls_per_interval set to 3 and the interval set to 1 second.
    # This matches with the limits configured for coinbase itself.
    with RateLimitedExecutionQueue(max_calls_per_interval=3, interval=1) as queue:
      self._download_with_queue(row_function, queue)

  @staticmethod
  def _write_row_to_csv(csv_writer, row: list):
    # Convert the time to an ISO timestamp.
    write_row = row[1:]
    write_row.insert(0, datetime.datetime.utcfromtimestamp(row[0]).isoformat())
    csv_writer.writerow(write_row)

  def download_and_write_to_file(self, output_filename: Text):
    """Downloads historical data and writes it to a file."""
    with open(output_filename, 'w', newline='') as csv_file:
      csv_writer = csv.writer(csv_file)
      csv_writer.writerow(HistoricalDownloader._COLUMN_HEADERS)
      self._download(partial(HistoricalDownloader._write_row_to_csv, csv_writer))

  @staticmethod
  def _write_row_to_dataframe(df: pd.DataFrame, row: list):
    timestamp = pd.Timestamp(datetime.datetime.utcfromtimestamp(row[0]).isoformat())
    logging.debug('writing {} to {}'.format(timestamp, row))
    df.loc[timestamp] = row

  def download_to_dataframe(self):
    """Downloads the historical data specified by this class to a dataframe."""
    df = pd.DataFrame(columns=HistoricalDownloader._COLUMN_HEADERS)
    self._download(partial(HistoricalDownloader._write_row_to_dataframe, df))
    df.set_index('time')
    return df

  def download_to_list(self):
    """Downloads the historical data in the form of a list, like it would be returned by Coinbase."""
    output = list()
    self._download(output.append)
    return output

  @staticmethod
  def _solve_required_calls(start_time: datetime.datetime, end_time: datetime.datetime, granularity: int):
    """Solves the calls that are required to get all date between start_time and end_time with the given granularity.

    Returns:
      A list of tuples representing the start and end times required to get all data between start_time and end_time.
    """
    total_interval = end_time - start_time
    # First solve the max delta for the interval (this is the largest delta we can use)
    max_delta = datetime.timedelta(seconds=HistoricalDownloader._MAX_CANDLES * granularity)
    # Special case, we can just make a single call and we will be under the maximum amount that can
    # be returned by the API.
    if total_interval < max_delta:
      return [(start_time, end_time)]
    required_calls = []
    current_time = start_time
    num_required_calls = math.floor(total_interval / max_delta)
    for _ in range(num_required_calls):
      new_end_time = current_time + max_delta
      required_calls.append((current_time, new_end_time))
      current_time = new_end_time
    # The last interval goes to the end time.
    required_calls.append((current_time, end_time))
    return required_calls

  @staticmethod
  def validate_granularity(granularity: Text):
    return granularity in list(HistoricalDownloader.TIMESLICE_MAPPINGS.keys()) + list(
      HistoricalDownloader.TIMESLICE_MAPPINGS.values())



import websocket
import json
import logging
import threading

from .util import LogLevel
from .coinbase_auth import CoinbaseAuth

# Special channels are sometimes sent by Coinbase, but cannot be subscribed to directly.
SPECIAL_CHANNELS = ['error',
                    # Do an action when the websocket is opened/closed.
                    # The parameters for these functions will be the websocket.
                    'open_websocket',
                    'close_websocket',
                    # All Messages.
                    'all_messages',
                    # Sent when new subscriptions are created. Docs: https://docs.pro.coinbase.com/#subscribe
                    'subscriptions',
                    # Used by Coinbase "level2" channel, allowing you to keep a snapshot of the order-book locally.
                    # Documentation: https://docs.pro.coinbase.com/#the-level2-channel
                    'snapshot',
                    'l2update',
                    ]
# Channels received with "full" subscription: https://docs.pro.coinbase.com/#the-full-channel
FULL_CHANNELS = ['open', 'received', 'match', 'change', 'activate']


# noinspection PyUnusedLocal
class CoinbaseWebsocket:
  PROD_ADDRESS = 'wss://ws-feed.pro.coinbase.com'
  SANDBOX_ADDRESS = 'wss://ws-feed-public.sandbox.pro.coinbase.com'

  def __init__(self, websocket_addr=PROD_ADDRESS,
               products_to_listen=None,
               channels_to_function=None,
               extra_channels=None,
               preparse_json=True,
               autostart=True,
               log_level=LogLevel.BASIC_MESSAGES,
               api_key=None, api_secret=None, passphrase=None):
    """Constructor for the CoinbaseWebsocket.

    Minimal Usage:
      # This will subscribe to the heartbeat channel and print all messages on the websocket.
      cbws = CoinbaseWebsocket(products_to_listen='BTC-USD',
                               channels_to_function={'heartbeat': lambda msg: print(msg) })

    You can also add a list of functions to run on each channel. Using add_channel_function will add
    a new function to the list.

    All three api_key, api_secret, and passphrase are required for authenticated websocket.

    Args:
      websocket_addr: The address to subscribe to. Default is prod, but you should use Sandbox for testing.
        Sandbox: 'wss://ws-feed-public.sandbox.pro.coinbase.com'
        Prod: 'wss://ws-feed-public.sandbox.pro.coinbase.com'
      products_to_listen: List of products to subscribe to when the socket is initially opened.
        Be warned, if you don't subscribe to something within 5 seconds of opening, the Websocket will be closed, you
        can call start_websocket to restart the websocket, but you're responsible for the threading that this class
        takes care of for you normally (if autostart is True)
      channels_to_function: Map of Channels to Functions, functions should take a single parameter (the message is
        parsed json, unless preparse_json is False).
        These functions will be called any time we get a message for a given channel.
        "Special" Channels:
          "error": Error Handler (gets json or string message)
          "close_websocket": Handle on-close (parameter is websocket)
          "open_websocket": Handle on-open (parameter is websocket)
          "full": Subscribing to this channel will create FULL_CHANNELS messages.
                  These messages might be useful for authenticated clients for confirming filling of orders.
      extra_channels: Extra channels to subscribe to without a function.
      preparse_json: (Default: True) Should we pass json to channels to function or simply the string?
      autostart: (Default: True) Start the websocket by default.
      log_level: (Default: ERROR_LOG) The LOG_LEVEL to use for this class, by default, will only report errors (using
        python logging api)
      api_key: (optional) API Key for Authenticated Websocket
      api_secret: (optional) API Secret for Authenticated Websocket
      passphrase: (optional) passphrase for authenticated websocket
    """
    if products_to_listen is None:
      products_to_listen = []
    if channels_to_function is None:
      channels_to_function = {}
    if extra_channels is None:
      extra_channels = []
    self.websocket_addr = websocket_addr
    self.products_to_listen = products_to_listen
    self.channels_to_function = channels_to_function
    self.extra_channels = extra_channels
    self.preparse_json = preparse_json
    self.log_level = log_level
    self.api_key = api_key
    self.api_secret = api_secret
    self.passphrase = passphrase
    self.ws = websocket.WebSocketApp(self.websocket_addr,
                                     on_message=lambda ws, msg: self.on_message(ws, msg),
                                     on_error=lambda ws, err: self.on_error(ws, err),
                                     on_close=lambda ws: self.on_close(ws),
                                     on_open=lambda ws: self.on_open(ws))
    if autostart:
      self.ws_thread = threading.Thread(target=self.start_websocket)
      self.ws_thread.start()

  def __del__(self):
    self.close_websocket()

  def start_websocket(self):
    self.ws.run_forever()

  def start_websocket_in_thread(self):
    self.ws_thread = threading.Thread(target=self.start_websocket)
    self.ws_thread.start()

  def close_websocket(self):
    self.ws.close()

  @staticmethod
  def make_subscribe(product_ids=None, channels=None,
                     api_key=None, api_secret=None, passphrase=None):
    if product_ids is None or channels is None:
      raise SyntaxError('Must specify channels and product_ids')
    subscribe_msg = {'type': 'subscribe', 'product_ids': product_ids, 'channels': channels}
    if api_key and api_secret and passphrase:
      subscribe_msg.update(CoinbaseAuth.get_websocket_verification(api_key, api_secret, passphrase))
    return json.dumps(subscribe_msg)

  def add_channel_function(self, channel, function, refresh_subscriptions=None):
    if channel in self.channels_to_function:
      functions = self.channels_to_function[channel]
      if isinstance(functions, list):
        # Already a list, just append.
        functions.append(function)
      else:
        # Not a list, make it a list.
        current_function = functions
        functions = [current_function, function]
    else:
      self.channels_to_function[channel] = [function]
      # This is a new channel, so force subscription update, if not set.
      if refresh_subscriptions is None:
        refresh_subscriptions = True
    # Force refresh subscriptions if not explicitly set.
    if refresh_subscriptions or refresh_subscriptions is None:
      self.subscribe()

  def add_product(self, product, refresh_subscriptions=True):
    if product not in self.products_to_listen:
      self.products_to_listen.append(product)
      if refresh_subscriptions:
        self.subscribe()

  def add_channel(self, channel, refresh_subscriptions=True):
    self.extra_channels.append(channel)

  def add_authentication(self, api_key, api_secret, passphrase):
    self.api_key = api_key
    self.api_secret = api_secret
    self.passphrase = passphrase
    self.subscribe()

  def subscribe(self):
    channels = [channel for channel in self.channels_to_function.keys() if
                channel not in SPECIAL_CHANNELS + FULL_CHANNELS] + self.extra_channels
    if self.log_level >= LogLevel.VERBOSE_LOG:
      logging.info("Subscribing to channels: {}".format(", ".join(channels)))
    # Make sure there are channels and products to listen to.
    if channels and self.products_to_listen:
      self.ws.send(CoinbaseWebsocket.make_subscribe(self.products_to_listen,
                                                    channels,
                                                    self.api_key, self.api_secret, self.passphrase))

  def on_open(self, ws):
    if self.log_level >= LogLevel.BASIC_MESSAGES:
      logging.info('Coinbase Websocket Connection ({})'.format(self.websocket_addr))
    # Subscribe to defaults.
    if self.products_to_listen and self.channels_to_function:
      self.subscribe()
    if 'open_websocket' in self.channels_to_function:
      self._execute_functions_on_message(ws, self._get_functions_as_list('open_websocket'))

  def on_error(self, ws, err):
    del ws  # We don't use this, but it's required by WebSocketApp.
    if self.log_level >= LogLevel.ERROR_LOG:
      logging.error('Error Received: {}'.format(err))
    self._call_message_functions(err)

  def on_message(self, ws, message):
    del ws  # We don't use this, but it's required by WebSocketApp.
    if self.log_level >= LogLevel.VERBOSE_LOG:
      logging.info('Message Received: {}'.format(message))
    self._call_message_functions(message)

  def on_close(self, ws):
    if self.log_level >= LogLevel.BASIC_MESSAGES:
      logging.info('Coinbase Websocket Disconnection ({})'.format(self.websocket_addr))
    if 'close_websocket' in self.channels_to_function:
      self._execute_functions_on_message(ws, self._get_functions_as_list('close_websocket'))

  def _call_message_functions(self, message):
    json_msg = json.loads(message)
    functions_to_execute = list()
    if 'type' in json_msg:
      if json_msg['type'] in self.channels_to_function:
        functions_to_execute.extend(self._get_functions_as_list(json_msg['type']))
      # Subscribe to "full" messages with "full" keyword.
      if 'full' in self.channels_to_function and json_msg['type'] in FULL_CHANNELS:
        functions_to_execute.extend(self._get_functions_as_list('full'))
      # Matches messages come on "match" messages.
      if 'matches' in self.channels_to_function and json_msg['type'] == 'match':
        functions_to_execute.extend(self._get_functions_as_list('matches'))
    if 'all_messages' in self.channels_to_function:
      functions_to_execute.extend(self._get_functions_as_list('all_messages'))
    self._execute_functions_on_message(message, functions_to_execute, json_msg)

  def _execute_functions_on_message(self, message, functions, json_msg=None):
    message_is_string = isinstance(message, str)
    if json_msg is None and message_is_string:
      json_msg = json.loads(message)
    for function in functions:
      function(json_msg if self.preparse_json and message_is_string else message)

  def _get_functions_as_list(self, message_type):
    functions_list = []
    functions = self.channels_to_function[message_type]
    if isinstance(functions, list):
      functions_list.extend(functions)
    else:
      functions_list.append(functions)
    return functions_list



from setuptools import setup

VERSION_NUMBER = '0.0.5'

setup(
  name='zcoinbase',
  version=VERSION_NUMBER,
  packages=['zcoinbase'],
  url='https://github.com/Zbot21/zcoinbase',
  download_url='https://github.com/Zbot21/zcoinbase/archive/v{}.tar.gz'.format(VERSION_NUMBER),
  license='MIT',
  author='Chris Bellis',
  author_email='chris@zbots.org',
  description='A Simple Coinbase Client for the Coinbase Pro API',
  install_requires=['requests',
                    'websocket-client',
                    'sortedcontainers',
                    'python-dateutil',
                    'pandas']
)
